local SQ = LibStub("AceAddon-3.0"):NewAddon("SmoothQuest", "AceEvent-3.0", "AceHook-3.0")

local questtags, tags = {}, {Elite = "+", Group = "g", Dungeon = "d", Raid = "r", PvP = "p", Heroic="d+", Daily = "\226\128\162"}
local objects_pattern = '^' .. QUEST_OBJECTS_FOUND:gsub('(%%.)', '(.+)') .. '$' --QUEST_OBJECTS_FOUND = "%s: %d/%d"
local monsters_pattern = '^' .. QUEST_MONSTERS_KILLED:gsub('(%%.)', '(.+)') .. '$' --QUEST_MONSTERS_KILLED = "%s slain: %d/%d"
local faction_pattern = '^' .. QUEST_FACTION_NEEDED:gsub('(%%.)', '(.+)') .. '$' --QUEST_FACTION_NEEDED = "%s: %s / %s"

local function ColorGradient(perc, ...)
	if perc >= 1 then
		local r, g, b = select(select('#', ...) - 2, ...)
		return r, g, b
	elseif perc <= 0 then
		local r, g, b = ...
		return r, g, b
	end

	local num = select('#', ...) / 3

	local segment, relperc = math.modf(perc*(num-1))
	local r1, g1, b1, r2, g2, b2 = select((segment*3)+1, ...)

	return r1 + (r2-r1)*relperc, g1 + (g2-g1)*relperc, b1 + (b2-b1)*relperc
end

local function rgb2hex( r, g, b )
	if type(r) == "table" then
		g = r.g
		b = r.b
		r = r.r
	end
	return string.format("%02x%02x%02x", r*255, g*255, b*255)
end


local function GetTaggedTitle(i)
	local name, level, tag, group, header, _, _, daily = GetQuestLogTitle(i)
	if header or not name then return end

	if not group or group == 0 then group = nil end
	return string.format("|cff%s[%s%s%s] %s|r", rgb2hex(GetDifficultyColor(level)), level, tag and tags[tag] or daily and tags.Daily or "", group or "", name)
end


function SQ:OnEnable()
	self:SecureHook("QuestLog_Update")
	self:SecureHook("QuestWatch_Update")
	self:RegisterEvent("QUEST_LOG_UPDATE")
	self:HookScript(GameTooltip, "OnTooltipSetItem")
	self:HookScript(GameTooltip, "OnTooltipSetUnit")	
	
	self:QUEST_LOG_UPDATE()
	self:QuestWatch_Update()
end

local items, mobs, progress = {}, {}, {}
local mcache, pcache = {}, {}

function SQ:OnTooltipSetUnit( tooltip, ... )
	local name = tooltip:GetUnit()
	if mobs[name] then
		if type(mobs[name]) == "table" then
			for k, it in ipairs(mobs[name]) do
				tooltip:AddDoubleLine(progress[it].q, progress[it].i..'/'..progress[it].n, 1, 1, 1, ColorGradient( progress[it].perc, 1,0,0, 1,1,0, 0,1,0))
			end
		else
			local it = mobs[name]
			tooltip:AddDoubleLine(progress[it].q, progress[it].i..'/'..progress[it].n, 1, 1, 1, ColorGradient( progress[it].perc, 1,0,0, 1,1,0, 0,1,0))
		end
	end
end

function SQ:OnTooltipSetItem( tooltip, ...)
	local name = tooltip:GetItem()
	if items[name] then
		local it = items[name]
		if progress[it] then
			tooltip:AddDoubleLine(progress[it].q, progress[it].i..'/'..progress[it].n, 1, 1, 1, ColorGradient( progress[it].perc, 1,0,0, 1,1,0, 0,1,0))
		end
	end
end

local function getPCache()
	local t = next(pcache)
	if t then
		pcache[t] = nil
	else t = {} end
	return t
end

local function getMCache()
	local t = next(mcache)
	if t then
		mcache[t] = nil
	else t = {} end
	return t
end

local function emptyAll()
	for k, v in pairs(items) do
		items[k] = nil
	end
	for k, v in pairs(progress) do
		if type(v) == "table" then
			pcache[v] = true
		end
		progress[k] = nil
	end
	for k, v in pairs(mobs) do
		if type(v) == "table" then
			mcache[v] = true
		end
		mobs[k] = nil
	end
end

-- Modified update from Quixote
function SQ:QUEST_LOG_UPDATE()
	emptyAll()
	local startingQuestLogSelection = GetQuestLogSelection()
	local numEntries, numQuests = GetNumQuestLogEntries()
	if numEntries > 0 then
		local questid
		for questid = 1, numEntries do
			SelectQuestLogEntry(questid) -- this is required for a few of the quest functions.
			local strQuestLogTitleText, strQuestLevel, strQuestTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily = GetQuestLogTitle(questid)
			if not isHeader then
				local strQuestDescription, strQuestObjectives = GetQuestLogQuestText(questid)
				--Many mods like to hook GetQuestLogTitle to add levels to the names.  This is annoying, so strip out the common format for it.
				if strQuestLogTitleText:match('^%[') then strQuestLogTitleText = strQuestLogTitleText:match("^%[[^%]]+%]%s?(.*)") end
				-- quest_names[strQuestLogTitleText] = questid
				-- self:Print( strQuestlogTitleText, strQuestObjectives )
				local numObj = GetNumQuestLeaderBoards(questid)
				if numObj > 0 then
					local ii
					for ii=1, GetNumQuestLeaderBoards(questid) do
						local ndesc
						local desc, qtype, done = GetQuestLogLeaderBoard(ii, questid)
						if qtype == 'item' or qtype == 'object' then --'object' for the leaderboard in Dousing the Flames of Protection; maybe others.
							ndesc,numItems,numNeeded = desc:match(objects_pattern)
							items[ndesc] = desc -- used for tooltips
							-- items[desc] = ("%s: %d/%d"):format(strQuestLogTitleText, numItems, numNeeded)
						elseif qtype == 'monster' then
							mobName,numItems,numNeeded = desc:match(monsters_pattern)
							if mobName == nil or numItems == nil or numNeeded == nil then
								--Sometimes we get objectives like "Find Mankrik's Wife: 0/1", which are listed as "monster".
								mobName, numItems, numNeeded = desc:match(objects_pattern)
							end
							ndesc = mobName
							if mobs[ndesc] then
								--Another quest also wants this mob!  Convert mobs[desc] to a table.
								if type(mobs[ndesc]) == 'string' then
									local s = mobs[ndesc]
									mobs[ndesc] = getMCache()
									mobs[ndesc][1] = s
								end
								local t = mobs[ndesc]
								t[#t+1] = desc -- ("%s: %d/%d"):format(strQuestLogTitleText, numItems, numNeeded)
							else
								mobs[ndesc] = desc -- ("%s: %d/%d"):format(strQuestLogTitleText, numItems, numNeeded)
							end
						elseif qtype == 'reputation' then
							ndesc,numItems,numNeeded = desc:match(faction_pattern)
						elseif qtype == 'event' then
							numNeeded = 1
							if done then
								numItems = 1
							else
								numItems = 0
							end
							ndesc = desc
						end
						numNeeded = tonumber(numNeeded)
						numItems = tonumber(numItems)
						if numNeeded > 0 then
							if not progress[desc] then
								progress[desc] = next(pcache) or {}
							end
							progress[desc].q = strQuestLogTitleText
							progress[desc].i = numItems
							progress[desc].n = numNeeded
							progress[desc].perc = numItems / numNeeded
						end
					end
				end
			end
			
			
		end
	end
	SelectQuestLogEntry(startingQuestLogSelection)
end


-- Add tags to the quest log, integral from QueLevel
function SQ:QuestLog_Update()
	local offset, numEntries, numQuests = FauxScrollFrame_GetOffset(QuestLogListScrollFrame), GetNumQuestLogEntries()

	for i=1,QUESTS_DISPLAYED do
		local qi = i + offset
		local questLogTitle, questTitleTag, questNormalText, questCheck = _G["QuestLogTitle"..i], _G["QuestLogTitle"..i.."Tag"], _G["QuestLogTitle"..i.."NormalText"], _G["QuestLogTitle"..i.."Check"]

		if qi <= numEntries then
			local _, _, tag, _, _, _, complete, daily = GetQuestLogTitle(qi)

			local title = GetTaggedTitle(qi)
			if title then
				questLogTitle:SetText("  "..title)
				QuestLogDummyText:SetText("  "..title)
			end

			if IsQuestWatched(qi) then
				questCheck:ClearAllPoints()
				questCheck:SetPoint("RIGHT", questLogTitle, "LEFT", 24, 0)
			end
			
			if tag or complete and complete ~= 0 or daily then
				if not complete then questTitleTag:SetText("") end
				local tempWidth = 275 - 15 - questTitleTag:GetWidth()
				local textWidth = math.min(QuestLogDummyText:GetWidth(), tempWidth)
				questNormalText:SetWidth(tempWidth)
			end
		end
	end
end

-- Add tags to the quest watcher, taken from QueLevel and modified for leaderboard color additions
function SQ:QuestWatch_Update()
	local questWatchMaxWidth, watchTextIndex = 0, 1

	for i=1,GetNumQuestWatches() do
		local qi = GetQuestIndexForWatch(i)
		if qi then
			local numObjectives = GetNumQuestLeaderBoards(qi)

			if numObjectives > 0 then
				local text = _G["QuestWatchLine"..watchTextIndex]
				text:SetText(GetTaggedTitle(qi))
				local tempWidth = text:GetWidth()
				questWatchMaxWidth = math.max(tempWidth, questWatchMaxWidth)
				local oldwatchText = watchTextIndex
				watchTextIndex = watchTextIndex + 1
				local text, type, finished
				for j=1,numObjectives do
					textt, ttype, finished = GetQuestLogLeaderBoard(j,qi)
					if progress[textt] then
						textt = string.format(" - |cff%s%s|r", rgb2hex( ColorGradient( progress[textt].perc, 1,0,0, 1,1,0, 0,1,0) ), textt )
						text = _G["QuestWatchLine"..watchTextIndex]
						text:SetText(textt)
					end
					watchTextIndex= watchTextIndex + 1
				end
				watchTextIndex = oldwatchText + numObjectives + 1
			end
		end
	end

	if watchTextIndex ~= 1 and QuestWatchFrame:GetWidth() < (questWatchMaxWidth + 10) then QuestWatchFrame:SetWidth(questWatchMaxWidth + 10) end
end

